import pandas as pd
import numpy as np 
from matplotlib import pyplot as plt 
import os,sys,random,time,uuid,re 
from datetime import datetime
from sklearn.metrics import accuracy_score, f1_score

from utils import convert_predict_string_to_dict,dict_to_string, _generate_past_question_correctness_concise, question_dict_to_string, question_dict_to_string_concise, _extract_llm_predict_correctness, _extract_llm_predict_reason, _generate_past_question_correctness_info, _generate_future_question_info, _generate_specific_eval_string, _store_log
from reflection_distillation import Experiment_Pipeline
from llm_api import _response_llm_gpt


class LLM_Predictor():
    def __init__(self,log_folder,print_log,openai_api_key):
        self.openai_api_key = openai_api_key
        self.print_log = print_log
        self.log_folder = log_folder
        self.log_iter_reflect = self.log_folder + '/iterative_reflect.txt'
        self.log_guide_predict = self.log_folder + '/guide_predict.txt'
        self.log_guide_predict_from_history = self.log_folder + '/guide_predict_from_history.txt'
     
        self.reflect_strategy = ""
        self.reflect_format = 'Your output format should be specific for each wrongly predicted question like << The reason why I make a wrong prediction for question xxx is that. Specifically, I have ignored the estimation of student learning status when learning the concept of xxx, which is related to question xxx... If I have a new chance to make new predictions, I will consider... >>'
        

    def ideal_predict(self,dataset_past_item,dataset_future_item,model_type):
        past_question_correctness_str = _generate_past_question_correctness_info(dataset_past_item)
        future_question_str = _generate_future_question_info(dataset_future_item)
        task_str = (
            'Your task is to predict the student answer correctness of several future questions, based on the student answer correctness of past questions, '
            +'as well as the potential correlation of concepts between past questions and future questions.\n\n'
        )
        format_str = (
            'Based on the info above, please predict whether this student will answer each of the future questions correctly or not.'
            +'Note that you should give response for each one of the future questions. Please give both prediction results with your reasons per future question.'
            +'Your output format should be: Question ID: Correct or Wrong, Reason: reason. \n'
            +'For example, Question 1: Correct, Reason: your reason 1; Question 2: Wrong, Reason: your reason 2; Question 3: Wrong, Reason: your reason 3.  \n'
        )
        
        initial_input_prompt = task_str + past_question_correctness_str + future_question_str + format_str

        message_sys = {"role": "system","content": "You are an intelligent assistant."}
        message_init_user = {"role": "user","content": initial_input_prompt}
        message_list_predict = [message_sys,message_init_user] 

        initial_response = _response_llm_gpt(message_list_predict,self.openai_api_key,model_type)
        future_correctness_predict_dict = _extract_llm_predict_correctness(initial_response,pattern = r"question\s*(\d+)\s*:\s*(correct|wrong)")
        future_correctness_reason_dict = _extract_llm_predict_reason(initial_response,pattern = r'question\s*(\d+)\s*:\s*(\w+)\s*,\s*reason\s*:\s*(.*?)\s*(?=;|\.)')

        input_past_prompt = past_question_correctness_str

        fu_question_list = list(set(dataset_future_item['question_id'])) 
        input_future_prompt_dict = {}
        for fq in fu_question_list:
            dataset_future_item_per = dataset_future_item[dataset_future_item['question_id']==fq]
            input_future_prompt_dict[fq] = _generate_future_question_info(dataset_future_item_per)
        llm_response_dict = future_correctness_reason_dict

        return initial_input_prompt,initial_response,future_correctness_predict_dict,input_past_prompt,input_future_prompt_dict,llm_response_dict

    
    def guide_predict(self,initial_input_prompt,initial_response,reflected_guidance,model_type):
        sys_prompt = "\n\nYou are an intelligent assistant."
        message_sys = {"role": "system","content": sys_prompt}
        message_init_user = {"role": "user","content": initial_input_prompt}
        message_init_response = {"role": "assistant","content": initial_response}
        guide_prompt = '\nHere are your reflections for your old predictions. '+reflected_guidance+'\n\nBased on the reflections, please ** REVISE ** your old predictions and make new predictions for all future questions.\n'
        message_list_predict = [message_sys,message_init_user,message_init_response,{"role": "user","content": guide_prompt}] 
        predict_str = _response_llm_gpt(message_list_predict,self.openai_api_key,model_type)
        future_predict_dict = _extract_llm_predict_correctness(predict_str,pattern = r"question\s*(\d+)\s*:\s*(correct|wrong)")
        _store_log(sys_prompt,self.log_guide_predict,color='green',print=self.print_log)
        _store_log(initial_input_prompt,self.log_guide_predict,color='green',print=self.print_log)
        _store_log(initial_response,self.log_guide_predict,color='green',print=self.print_log)
        _store_log(guide_prompt,self.log_guide_predict,color='green',print=self.print_log)
        _store_log(predict_str,self.log_guide_predict,color='green',print=self.print_log)
        return future_predict_dict,predict_str

    def guide_predict_from_history(self,reflected_history,test_data_past,model_type):
        past_correctness_str = _generate_past_question_correctness_concise(test_data_past)
        guide_prompt = (
            f'\nBased on the reflections above, now here is a new student whose correctness of the same past questions is: {past_correctness_str}. '
            +'please make new reflections and new predictions for correctness of the same future questions of this specific NEW student.\n'
            +'Your reflection output should be put into special symbol << >>. One example is << My new reflection for the new student is ... >>. '
            'Your new prediction format for the New student should be the same as before, i.e. Question ID: Correct or Wrong, Reason: reason. \n'
        )
        message_list_predict = reflected_history + [{"role": "user","content": guide_prompt}] 
        predict_str = _response_llm_gpt(message_list_predict,self.openai_api_key,model_type)
        future_predict_dict = _extract_llm_predict_correctness(predict_str,pattern = r"question\s*(\d+)\s*:\s*(correct|wrong)")
        for message_item in reflected_history:
            _store_log('\n'+message_item["role"]+': '+message_item["content"],self.log_guide_predict_from_history,color='green',print=self.print_log)
        _store_log('\nllm response: '+predict_str,self.log_guide_predict_from_history,color='blue',print=self.print_log)
        return future_predict_dict,predict_str

    def iterative_reflect(self,initial_input_prompt,initial_response,future_correctness_predict_dict,dataset_future,iter_threshold = 5,llm_model_type = '3'):
        future_question_id_list = list(set(dataset_future['question_id']))
        future_question_id_list.sort()

        future_label_dict = dataset_future.set_index('question_id')['correctness'].to_dict()
        assert len(future_question_id_list) == len(future_label_dict)
        future_label_str = question_dict_to_string(future_label_dict)

        future_accuracy_init_list = [1 if future_correctness_predict_dict[future_question_id] == future_label_dict[future_question_id] else 0 for future_question_id in future_question_id_list]
        future_accuracy_init = np.mean(future_accuracy_init_list)

        specific_eval_str = _generate_specific_eval_string(future_correctness_predict_dict,future_label_dict)

        init_reflect_str = (
            f'\n\nHowever, the real labelled answering correctness for the student is as follows: {future_label_str}.\n\n'
            +specific_eval_str+'\n\n'
            +'\nPlease reflect based on the difference between your predictions and the labels. '
            +'\n Specifically, for each future question which is wrongly predicted, reflect and give the reason why you make the wrong prediction.'
            # +'As the outcome of your reflection, you should output a kind of suggestions that can guide future predictions to be more accurate. '
            # +'To test your reflection, I will apply your output guidance suggestions to a new prediction model to see potential improvement.'
            +self.reflect_strategy
            +self.reflect_format
            # +'\n\nPlease only output your reflections to guide my model to perform the thinking and reasoning during the prediction process. '
            # +'Do not give any other info such as your rationales to do so.'
            # +'MAKE SURE your reflection guidance is specific to specific questions and related course concetps. It should NOT be simply high-level sentences!'
            # +'For example, you can say << when you reason the prediction process, you should consider how question A is related to question B. But you should also consider whether students may have good knowledge during learning concepts related to question C from how students perform in question D to estimate if students are learning well or not... >>.'
        )

        sys_prompt = "You are an intelligent assistant."
        message_sys = {"role": "system","content": sys_prompt}
        message_init_user = {"role": "user","content": initial_input_prompt}
        message_init_response = {"role": "assistant","content": initial_response}
        message_init_reflect = {"role": "user","content": init_reflect_str}
        message_list_reflect = [message_sys,message_init_user,message_init_response,message_init_reflect]

        _store_log(sys_prompt,self.log_iter_reflect,color='red',print=self.print_log)
        _store_log(initial_input_prompt,self.log_iter_reflect,color='red',print=self.print_log)
        _store_log(initial_response,self.log_iter_reflect,color='red',print=self.print_log)
        _store_log(init_reflect_str,self.log_iter_reflect,color='red',print=self.print_log)

        future_accuracy_dict, reflected_guidance_dict, future_predict_dict_iteration, llm_predict_str_dict = {}, {}, {}, {}

        for iteration_id in range(iter_threshold):
            reflected_guidance = _response_llm_gpt(message_list_reflect,self.openai_api_key,llm_model_type)
            _store_log(reflected_guidance,self.log_iter_reflect,color='red',print=self.print_log)
            future_predict_dict,llm_predict_str = self.guide_predict(initial_input_prompt,initial_response,reflected_guidance,llm_model_type)
            if len(future_label_dict)!=len(future_predict_dict) or list(future_label_dict.keys())!=list(future_predict_dict.keys()): continue 

            future_accuracy_list = [1 if future_predict_dict[future_question_id] == future_label_dict[future_question_id] else 0 for future_question_id in future_question_id_list]
            future_accuracy_avg = np.mean(future_accuracy_list)
            future_accuracy_dict[iteration_id] = future_accuracy_avg
            llm_predict_str_dict[iteration_id] = llm_predict_str
            future_predict_dict_iteration[iteration_id] = future_predict_dict
            reflected_guidance_dict[iteration_id] = reflected_guidance
            future_predict_str = question_dict_to_string(future_predict_dict)
            specific_eval_str = _generate_specific_eval_string(future_predict_dict,future_label_dict) + '\n\nBelow are the new predictions and reasoning based on your previouly provided reflections.\n\n' + llm_predict_str
            if future_accuracy_avg > future_accuracy_init:
                eval_str = f'I tried to use your reflection to let the model make the prediction again. The prediction accuracy is indeed improved.\n'
                direction_str = "\n\nPlease continue reflection to see whether we can further improve the prediction accuracy.\n\n "
            elif future_accuracy_avg < future_accuracy_init:
                eval_str = f'I tried to use your reflection to let the model make the prediction again. However, the prediction accuracy is worse than before.\n'
                direction_str = "\n\nPlease continue reflection in another direction to see whether we can further improve the prediction accuracy.\n\n "
            else:
                eval_str = f'I tried to use your reflection to let the model make the prediction again. However, the prediction accuracy is the same.\n'
                direction_str = "\n\nPlease continue reflection in another direction to see whether we can further improve the prediction accuracy.\n\n "
            
            continue_reflect_prompt = (
                eval_str+specific_eval_str+direction_str+self.reflect_format
                # +'Again, please only output your reflections to guide my model to perform the thinking and reasoning during the prediction process. '
                # +'Again, your output format should be specific for each wrongly predicted question like << The reason why I make a wrong prediction for question xxx is that... If I have a new chance to make new predictions, I will consider... >>'
                # +'For example, you can say << when you reason the prediction process, you should consider xxx >>.'
                # +'Do not give any other info such as your rationales to do so.'
            )
            message_list_reflect.append({"role": "assistant","content": reflected_guidance})
            message_list_reflect.append({"role": "user","content": continue_reflect_prompt})
            _store_log(continue_reflect_prompt,self.log_iter_reflect,color='red',print=self.print_log)
            if future_accuracy_avg == 1: break
        
        if len(future_accuracy_dict) == 0:
            return None,None,None,None,None,None
        best_iteration = max(future_accuracy_dict, key=future_accuracy_dict.get)
        return best_iteration,reflected_guidance_dict,future_accuracy_init,future_accuracy_dict,future_predict_dict_iteration,llm_predict_str_dict


class Transferable_Iterative_Reflection():
    def __init__(self,llm_model_type,dataset_path,log_folder,print_log,openai_api_key):
        self.llm_model_type = llm_model_type
        self.dataset_path = dataset_path
        self.reflection_database_path = log_folder+'/reflection_database.csv'
        self.predict_post_path = log_folder+'/predict_post.csv'
        self.predict_post_detail_path = log_folder+'/predict_post_detail.csv'
        self.result_path = log_folder+'/result.csv'
        self.result_item_path = log_folder+'/result_item.csv'
        self.llm_predictor = LLM_Predictor(log_folder,print_log,openai_api_key)


    def dataset_prepare(self,datasample_flag=True,reflect_sample_ratio=0.2,random_seed=4):
        self.dataset = pd.read_csv(self.dataset_path,sep='\t')
        if datasample_flag == True:
            self.dataset_train = self.dataset[self.dataset['data_type']=='train']
            np.random.seed(random_seed)
            self.reflect_students = np.random.choice(self.dataset_train['student_id'].unique(),size=int(reflect_sample_ratio * self.dataset_train['student_id'].nunique()),replace=False)
            self.dataset_reflect = self.dataset_train[self.dataset_train['student_id'].isin(self.reflect_students)]
        else:
            self.reflect_students = list(set(self.dataset['student_id']))
            self.dataset_reflect = self.dataset.copy()

    def dataset_prepare_for_prediction(self):
        self.dataset = pd.read_csv(self.dataset_path,sep='\t')
        self.dataset_train = self.dataset[self.dataset['data_type']=='train']
        self.dataset_predict = self.dataset[self.dataset['data_type']=='test']

    def check_exist_result(self):
        header_result_item = ['uid','data_type','student_id','question_id','llm_correctness','future_correctness_predict','future_label','input_past_prompt','input_future_prompt','initial_response','best_reflection']
        header_result = ['data_type','student_id','iteration_id','best','future_accuracy_init','future_accuracy_iteration','future_label','future_predict_init','future_predict_iteration','reflection']
        header_database = header_result + ['input_prompt','initial_response']
        header_result_item_str = '\t'.join(header_result_item)+'\n'
        header_result_str = '\t'.join(header_result)+'\n'
        header_database_str = '\t'.join(header_database)+'\n'
        if os.path.exists(self.reflection_database_path):
            with open(self.reflection_database_path, mode='r') as file1:
                first_line = file1.readline().strip()  # Read and strip the first line
                if not first_line:
                    with open(self.reflection_database_path, "a+") as file1:
                        file1.write(header_database_str)
                    student_list_exist = []
                else:
                    existing_reflection_database = pd.read_csv(self.reflection_database_path,sep='\t')
                    student_list_exist = list(set(existing_reflection_database['student_id']))
        else:
            with open(self.reflection_database_path, "a+") as file1:
                file1.write(header_database_str)
            student_list_exist = []

        if os.path.exists(self.result_path):
            with open(self.result_path, mode='r') as file1:
                first_line = file1.readline().strip()  # Read and strip the first line
                if not first_line:
                    with open(self.result_path, "a+") as file1:
                        file1.write(header_result_str)
        else:
            with open(self.result_path, "a+") as file1:
                file1.write(header_result_str)

        if os.path.exists(self.result_item_path):
            with open(self.result_item_path, mode='r') as file1:
                first_line = file1.readline().strip()  # Read and strip the first line
                if not first_line:
                    with open(self.result_item_path, "a+") as file1:
                        file1.write(header_result_item_str)
        else:
            with open(self.result_item_path, "a+") as file1:
                file1.write(header_result_item_str)
        
        return student_list_exist

    def run_reflect(self,student_num,iter_threshold):
        student_list_exist = self.check_exist_result()
        
        student_list = self.reflect_students[0:student_num]
        student_list.sort()
        
        for student_id in student_list:
            if student_id in student_list_exist: continue 
            print(f'running for student: {student_id}')
            dataset_student = self.dataset_reflect[self.dataset_reflect['student_id']==student_id]
            # TODO: ISSUE! The sample_uid is not unique since one student has several questions!!!!!!
            
            dataset_past = dataset_student[dataset_student['xy_type']=='past']
            dataset_future = dataset_student[dataset_student['xy_type']=='future']
            initial_input_prompt,initial_response,future_correctness_predict_dict,input_past_prompt,input_future_prompt_dict,llm_response_dict = self.llm_predictor.ideal_predict(dataset_past,dataset_future,self.llm_model_type)
            future_label_dict = dataset_future.set_index('question_id')['correctness'].to_dict()
            if len(future_label_dict)!=len(future_correctness_predict_dict) or list(future_label_dict.keys())!=list(future_correctness_predict_dict.keys()): continue 
            best_iteration,reflected_guidance_dict,future_accuracy_init,future_accuracy_dict,future_predict_dict_iteration,llm_predict_str_dict = self.llm_predictor.iterative_reflect(initial_input_prompt,initial_response,future_correctness_predict_dict,dataset_future,iter_threshold,self.llm_model_type)
            if best_iteration == None: continue
            best_reflection = reflected_guidance_dict[best_iteration]
            # if iter_threshold == -1:
            question_list = list(future_label_dict.keys())
            question_list.sort()
            for question_id in question_list:
                dataset_item = dataset_future[dataset_future['question_id']==question_id]
                sample_uid = dataset_item['uid'].values[0]
                data_type = dataset_item['data_type'].values[0]
                future_correctness_predict_item = future_correctness_predict_dict[question_id]
                future_label_item = future_label_dict[question_id]
                llm_correctness = 1 if future_correctness_predict_item == future_label_item else 0
                with open(self.result_item_path, "a+") as file1:
                    file1.write(str(sample_uid)+'\t'+data_type+'\t'+str(student_id)+'\t'+str(question_id)+'\t'+str(llm_correctness)+'\t'+str(future_correctness_predict_item)+'\t'+str(future_label_item)+'\t'+input_past_prompt.replace('\n','. ').replace('\t','. ')+'\t'+input_future_prompt_dict[question_id].replace('\n','. ').replace('\t','. ')+'\t'+llm_response_dict[question_id].replace('\n','. ').replace('\t','. ')+'\t'+best_reflection.replace('\n','. ').replace('\t','. ')+'\n')
                # continue
            iteration_list = list(reflected_guidance_dict.keys())
            iteration_list.sort()
            data_type = dataset_student['data_type'].values[0]
            for iteration_id in iteration_list:
                reflected_guidance = reflected_guidance_dict[iteration_id]
                best_flag = 1 if iteration_id == best_iteration else 0
                with open(self.result_path, "a+") as file1:
                    file1.write(data_type+'\t'+str(student_id)+'\t'+str(iteration_id)+'\t'+str(best_flag)+'\t'+str(future_accuracy_init)+'\t'+str(future_accuracy_dict[iteration_id])+'\t'+question_dict_to_string_concise(future_label_dict)+'\t'+question_dict_to_string_concise(future_correctness_predict_dict)+'\t'+question_dict_to_string_concise(future_predict_dict_iteration[iteration_id])+'\t'+reflected_guidance.replace('\n','. ').replace('\t','. ')+'\n')
                with open(self.reflection_database_path, "a+") as file1:
                    file1.write(data_type+'\t'+str(student_id)+'\t'+str(iteration_id)+'\t'+str(best_flag)+'\t'+str(future_accuracy_init)+'\t'+str(future_accuracy_dict[iteration_id])+'\t'+question_dict_to_string_concise(future_label_dict)+'\t'+question_dict_to_string_concise(future_correctness_predict_dict)+'\t'+question_dict_to_string_concise(future_predict_dict_iteration[iteration_id])+'\t'+reflected_guidance.replace('\n','. ').replace('\t','. ')+'\t'+initial_input_prompt.replace('\n','. ').replace('\t','. ')+'\t'+initial_response.replace('\n','. ').replace('\t','. ')+'\n')

    def check_exist_result_for_predict(self):
        header_result = ['data_type','student_id','future_accuracy_init','future_accuracy_post','future_label','future_predict_init','future_predict_post','reflection']
        header_database = header_result + ['input_prompt','initial_response']
        header_result_str = '\t'.join(header_result)+'\n'
        header_database_str = '\t'.join(header_database)+'\n'
        if os.path.exists(self.predict_post_detail_path):
            with open(self.predict_post_detail_path, mode='r') as file1:
                first_line = file1.readline().strip()  # Read and strip the first line
                if not first_line:
                    with open(self.predict_post_detail_path, "a+") as file1:
                        file1.write(header_database_str)
                    student_list_exist = []
                else:
                    existing_reflection_database = pd.read_csv(self.predict_post_detail_path,sep='\t')
                    student_list_exist = list(set(existing_reflection_database['student_id']))
        else:
            with open(self.predict_post_detail_path, "a+") as file1:
                file1.write(header_database_str)
            student_list_exist = []

        if os.path.exists(self.predict_post_path):
            with open(self.predict_post_path, mode='r') as file1:
                first_line = file1.readline().strip()  # Read and strip the first line
                if not first_line:
                    with open(self.predict_post_path, "a+") as file1:
                        file1.write(header_result_str)
        else:
            with open(self.predict_post_path, "a+") as file1:
                file1.write(header_result_str)

        
        return student_list_exist


    def run_predict_distill(self,student_num,distill_model_folder,dataset_raw_path):
        self.dataset_prepare_for_prediction()

        student_list_exist = self.check_exist_result_for_predict()

        distill_pipeline = Experiment_Pipeline(max_length=512, log_folder=distill_model_folder, dataset_raw_path=dataset_raw_path, load_model_type='last')
        
        student_list_predict = list(set(self.dataset_predict['student_id']))
        student_list_predict.sort()
        student_list_predict = student_list_predict[0:student_num]
        
        for student_id in student_list_predict:
            if student_id in student_list_exist: continue 
            print(f'running for student: {student_id}')
            dataset_student = self.dataset_predict[self.dataset_predict['student_id']==student_id]
            dataset_past = dataset_student[dataset_student['xy_type']=='past']
            dataset_future = dataset_student[dataset_student['xy_type']=='future']
            initial_input_prompt,initial_response,future_correctness_predict_dict,input_past_prompt,input_future_prompt_dict,llm_response_dict = self.llm_predictor.ideal_predict(dataset_past,dataset_future,self.llm_model_type)
            future_label_dict = dataset_future.set_index('question_id')['correctness'].to_dict()
            if len(future_label_dict)!=len(future_correctness_predict_dict) or list(future_label_dict.keys())!=list(future_correctness_predict_dict.keys()): continue 
            
            reflected_guidance = distill_pipeline.model_predict(initial_input_prompt, initial_response)
            print('reflected_guidance: ',reflected_guidance)
            future_predict_dict,llm_predict_str = self.llm_predictor.guide_predict(initial_input_prompt,initial_response,reflected_guidance,self.llm_model_type)
            
            data_type = dataset_student['data_type'].values[0]

            future_question_id_list = list(set(dataset_future['question_id']))
            future_question_id_list.sort()

            future_label_dict = dataset_future.set_index('question_id')['correctness'].to_dict()
            assert len(future_question_id_list) == len(future_label_dict)
            future_label_str = question_dict_to_string(future_label_dict)

            future_accuracy_init_list = [1 if future_correctness_predict_dict[future_question_id] == future_label_dict[future_question_id] else 0 for future_question_id in future_question_id_list]
            future_accuracy_init = np.mean(future_accuracy_init_list)

            future_accuracy_predict_list = [1 if future_predict_dict[future_question_id] == future_label_dict[future_question_id] else 0 for future_question_id in future_question_id_list]
            future_accuracy_post = np.mean(future_accuracy_predict_list)

            with open(self.predict_post_path, "a+") as file1:
                file1.write(data_type+'\t'+str(student_id)+'\t'+str(future_accuracy_init)+'\t'+str(future_accuracy_post)+'\t'+question_dict_to_string_concise(future_label_dict)+'\t'+question_dict_to_string_concise(future_correctness_predict_dict)+'\t'+question_dict_to_string_concise(future_predict_dict)+'\t'+reflected_guidance.replace('\n','. ').replace('\t','. ')+'\n')
            with open(self.predict_post_detail_path, "a+") as file1:
                file1.write(data_type+'\t'+str(student_id)+'\t'+str(future_accuracy_init)+'\t'+str(future_accuracy_post)+'\t'+question_dict_to_string_concise(future_label_dict)+'\t'+question_dict_to_string_concise(future_correctness_predict_dict)+'\t'+question_dict_to_string_concise(future_predict_dict)+'\t'+reflected_guidance.replace('\n','. ').replace('\t','. ')+'\t'+initial_input_prompt.replace('\n','. ').replace('\t','. ')+'\t'+initial_response.replace('\n','. ').replace('\t','. ')+'\n')

    def check_exist_result_for_predict_reuse_reflect(self):
        header_result = ['data_type','student_id','future_accuracy_init','future_accuracy_post','future_label','future_predict_init','future_predict_post']
        header_result_detail = ['uid','data_type','student_id','question_id','future_label','input_past_prompt','input_future_prompt','initial_response','best_reflection']
        header_result_str = '\t'.join(header_result)+'\n'
        header_result_detail_str = '\t'.join(header_result_detail)+'\n'
        if os.path.exists(self.predict_post_path):
            with open(self.predict_post_path, mode='r') as file1:
                first_line = file1.readline().strip()  # Read and strip the first line
                if not first_line:
                    with open(self.predict_post_path, "a+") as file1:
                        file1.write(header_result_str)
                    student_list_exist = []
                else:
                    existing_reflection_database = pd.read_csv(self.predict_post_path,sep='\t')
                    student_list_exist = list(set(existing_reflection_database['student_id']))
        else:
            with open(self.predict_post_path, "a+") as file1:
                file1.write(header_result_str)
            student_list_exist = []

        if os.path.exists(self.predict_post_detail_path):
            with open(self.predict_post_detail_path, mode='r') as file1:
                first_line = file1.readline().strip()  # Read and strip the first line
                if not first_line:
                    with open(self.predict_post_detail_path, "a+") as file1:
                        file1.write(header_result_detail_str)
        else:
            with open(self.predict_post_detail_path, "a+") as file1:
                file1.write(header_result_detail_str)

        return student_list_exist

    def generate_reflection_history(self,course_name,example_student_num_per_course,random_seed=4,iter_threshold=3):
        dataset_train_course = self.dataset_train[self.dataset_train['course_name']==course_name]
        np.random.seed(random_seed)
        example_students = np.random.choice(dataset_train_course['student_id'].unique(),size=example_student_num_per_course,replace=False)
        example_students.sort()
        dataset_reflect_example = dataset_train_course[dataset_train_course['student_id'].isin(example_students)]

        sys_prompt = "\nYou are an intelligent assistant."
        message_sys = {"role": "system","content": sys_prompt}

        for s,student_id in enumerate(example_students):
            print(f'reflection for student: {student_id}')
            dataset_student = dataset_train_course[dataset_train_course['student_id']==student_id]
            dataset_past = dataset_student[dataset_student['xy_type']=='past']
            dataset_future = dataset_student[dataset_student['xy_type']=='future']
            initial_input_prompt,initial_response,future_correctness_predict_dict,input_past_prompt,input_future_prompt_dict,llm_response_dict = self.llm_predictor.ideal_predict(dataset_past,dataset_future,self.llm_model_type)
            future_label_dict = dataset_future.set_index('question_id')['correctness'].to_dict()
            if len(future_label_dict)!=len(future_correctness_predict_dict) or list(future_label_dict.keys())!=list(future_correctness_predict_dict.keys()): continue 
            
            best_iteration,reflected_guidance_dict,future_accuracy_init,future_accuracy_dict,future_predict_dict_iteration,llm_predict_str_dict = self.llm_predictor.iterative_reflect(initial_input_prompt,initial_response,future_correctness_predict_dict,dataset_future,iter_threshold,self.llm_model_type)

            if s == 0:
                message_init_user = {"role": "user","content": initial_input_prompt}
                message_init_response = {"role": "assistant","content": initial_response}
                reflection_history_list = [message_sys,message_init_user,message_init_response]

                reflection_history_list.append({"role": "user","content": '\n\nHere are the reflections based on the difference between your initial prediction and labels.\n' + reflected_guidance_dict[best_iteration]})
                reflection_history_list.append({"role": "assistant","content": llm_predict_str_dict[best_iteration]})
            else:
                past_correctness_str = _generate_past_question_correctness_concise(dataset_past)
                guide_prompt = f'\nBased on the reflections above, now here is a NEW student whose correctness of the same past questions is: {past_correctness_str}. please make NEW reflections and NEW predictions for correctness of the same future questions of this NEW student.\n'

                reflection_history_list.append({"role": "user","content": guide_prompt})
                reflection_history_list.append({"role": "assistant","content": '\nHere are my reflections for the specific new student where I may make fault predictions before. ' + reflected_guidance_dict[best_iteration] + '.\n\nHere are my new revised predictions: \n' + llm_predict_str_dict[best_iteration]})

        return reflection_history_list


    def run_predict_reuse_reflect(self,example_student_num_per_course,test_student_num_per_course,random_seed,iter_threshold):
        self.dataset_prepare_for_prediction()

        student_list_exist = self.check_exist_result_for_predict_reuse_reflect()

        course_id_list = list(set(self.dataset_predict['course_name']))
        course_id_list.sort()

        for course_id in course_id_list:
            data_pred_course = self.dataset_predict[self.dataset_predict['course_name']==course_id]

            student_list_predict = list(set(data_pred_course['student_id']))
            student_list_predict.sort()
            student_list_predict = student_list_predict[0:test_student_num_per_course]

            reflection_history_list = self.generate_reflection_history(course_id,example_student_num_per_course,random_seed,iter_threshold)
        
            for student_id in student_list_predict:
                if student_id in student_list_exist: continue 
                print(f'running for student: {student_id}')
                dataset_student = data_pred_course[data_pred_course['student_id']==student_id]
                dataset_past = dataset_student[dataset_student['xy_type']=='past']
                dataset_future = dataset_student[dataset_student['xy_type']=='future']
                initial_input_prompt,initial_response,future_correctness_predict_dict,input_past_prompt,input_future_prompt_dict,llm_response_dict = self.llm_predictor.ideal_predict(dataset_past,dataset_future,self.llm_model_type)
                future_label_dict = dataset_future.set_index('question_id')['correctness'].to_dict()
                if len(future_label_dict)!=len(future_correctness_predict_dict) or list(future_label_dict.keys())!=list(future_correctness_predict_dict.keys()): continue 
                
                future_predict_dict,llm_predict_str = self.llm_predictor.guide_predict_from_history(reflection_history_list,dataset_past,self.llm_model_type)
                
                data_type = dataset_student['data_type'].values[0]

                future_question_id_list = list(set(dataset_future['question_id']))
                future_question_id_list.sort()

                assert len(future_question_id_list) == len(future_label_dict)

                if len(future_predict_dict) != len(future_question_id_list): continue

                future_accuracy_init_list = [1 if future_correctness_predict_dict[future_question_id] == future_label_dict[future_question_id] else 0 for future_question_id in future_question_id_list]
                future_accuracy_init = np.mean(future_accuracy_init_list)

                future_accuracy_predict_list = [1 if future_predict_dict[future_question_id] == future_label_dict[future_question_id] else 0 for future_question_id in future_question_id_list]
                future_accuracy_post = np.mean(future_accuracy_predict_list)

                reflect_matches = re.findall(r'<<(.*?)>>', llm_predict_str)
                best_reflection = reflect_matches[0] 

                question_list = list(future_label_dict.keys())
                question_list.sort()
                for question_id in question_list:
                    dataset_item = dataset_future[dataset_future['question_id']==question_id]
                    sample_uid = dataset_item['uid'].values[0]
                    data_type = dataset_item['data_type'].values[0]
                    future_label_item = future_label_dict[question_id]
                    with open(self.predict_post_detail_path, "a+") as file1:
                        # ['uid','data_type','student_id','question_id','input_past_prompt','input_future_prompt','initial_response','best_reflection']
                        file1.write(str(sample_uid)+'\t'+data_type+'\t'+str(student_id)+'\t'+str(question_id)+'\t'+str(future_label_item)+'\t'+input_past_prompt.replace('\n','. ').replace('\t','. ')+'\t'+input_future_prompt_dict[question_id].replace('\n','. ').replace('\t','. ')+'\t'+llm_response_dict[question_id].replace('\n','. ').replace('\t','. ')+'\t'+best_reflection.replace('\n','. ').replace('\t','. ')+'\n')


                with open(self.predict_post_path, "a+") as file1:
                    file1.write(data_type+'\t'+str(student_id)+'\t'+str(future_accuracy_init)+'\t'+str(future_accuracy_post)+'\t'+question_dict_to_string_concise(future_label_dict)+'\t'+question_dict_to_string_concise(future_correctness_predict_dict)+'\t'+question_dict_to_string_concise(future_predict_dict)+'\n')

 





from .lm_classifier import LM_Classifier_Pipeline
import pandas as pd 


def run_exp_bertKT():
    dataset_path = 'data.csv'
    log_folder = './test/'

    dataframe = pd.read_csv(dataset_path,sep='\t')
    dataframe_train = dataframe[dataframe['data_type']=='train'].iloc[0:100]
    dataframe_valid = dataframe[dataframe['data_type']=='test'].iloc[0:50]
    dataframe_test = dataframe[dataframe['data_type']=='test'].iloc[0:50]
    
    experiment_pipeline = Experiment_Pipeline(512,log_folder,load_model_type='none',num_labels=3,lr=5e-7)
    experiment_pipeline.dataset_prepare(dataframe_train,dataframe_valid,dataframe_test,input_column_list = ['past_question_answer_prompt','future_question_prompt'],label_column = 'correctness',trim_type = 'longest_first',batch_size = 16)
    experiment_pipeline.model_train(epochs=3,vis=True)
    experiment_pipeline.model_eval(eval_mode='test')

    experiment_pipeline = Experiment_Pipeline(512,log_folder,load_model_type='best',num_labels=3,lr=5e-7)
    experiment_pipeline.dataset_prepare(dataframe_train,dataframe_valid,dataframe_test,input_column_list = ['past_question_answer_prompt','future_question_prompt'],label_column = 'correctness',trim_type = 'longest_first',batch_size = 16)
    experiment_pipeline.model_eval(eval_mode='test')

    experiment_pipeline = Experiment_Pipeline(512,log_folder,load_model_type='best',num_labels=3,lr=5e-7)
    all_preds, accuracy, f1 = experiment_pipeline.model_predict(dataframe_test, input_column_list = ['past_question_answer_prompt','future_question_prompt'], label_column = None, trim_type = 'longest_first')
    print(all_preds)
    print(accuracy, f1)
run_exp_bertKT()



demo_survey = {
    "q1": {
        "type": "text-entry",
        "content": "What is your name?",
        "action": "text entry",
        "instruction": "Output your answers with reasons within 20 words."
    },
    "q2": {
        "type": "single-choice",
        "content": "What is your favorite fruit?",
        "action": "A. apple, B. banana, C. lemon",
        "instruction": "Output your single choice directly without reasons. Your output format should exactly like: choice: [choice]."
    },
    "q3": {
        "type": "multiple-choice",
        "content": "What do you eat everyday?",
        "action": "A. noodles, B. rice, C. meat",
        "instruction": "Output your multiple choices directly with reasons. Your output format should be exactly like: choice: [choice], reason: []. For example, if you want to select A and C, your output should be: choice: A,C, reason: [my reasons...]."
    }
}

demo_user_persona = {
    "u1": {
        "age": "24 years old",
        "job": "engineer"
    },
    "u2": {
        "education": "high school"
    },
    "u3": {
        "name": "Tom",
        "major": "Computer Science",
        "degree": "Master"
    }
}


import os, sys, time, json, math 
import numpy as np 
import pandas as pd 

import openai
from openai import OpenAI

def _response_llm_gpt(message_list,openai_api_key,model_name="gpt-3.5-turbo-0125",timeout=120,temperature=0,max_tokens=3000,print_progress=False):
    openai.api_key = openai_api_key
    client = OpenAI(api_key = openai.api_key)
    response = ''
    except_waiting_time = 1
    max_waiting_time = 32
    current_sleep_time = 0.5
    
    start_time = time.time()
    while response == '':
        try:
            completion = client.chat.completions.create(
                model=model_name, # gpt-4o, gpt-4
                messages=message_list,
                temperature=temperature,
                timeout = timeout,
                max_tokens=max_tokens
            )
                
            response = completion.choices[0].message.content
        except Exception as e:
            print(e)
            time.sleep(current_sleep_time)
            if except_waiting_time < max_waiting_time:
                except_waiting_time *= 2
            current_sleep_time = np.random.randint(0, except_waiting_time-1)
    end_time = time.time()   
    if print_progress:
        print('llm response time: ',end_time-start_time)     
    return response

class LLM_Respond_GPT_Class():
    def __init__(self,openai_api_key,model_name="gpt-3.5-turbo-0125",timeout=120,temperature=0,max_tokens=3000):
        self.openai_api_key = openai_api_key
        self.model_name = model_name
        self.timeout = timeout
        self.temperature = temperature
        self.max_tokens = max_tokens

    def run_response(self,message_list,print_progress):
        return _response_llm_gpt(message_list,self.openai_api_key,self.model_name,self.timeout,self.temperature,self.max_tokens,print_progress)







class Node_Agent():
    def __init__(self,llm_response_class):
        self.llm_response_class = llm_response_class


    def run_respond(conversation_dict,print_progress,save_log_txt_path=None):
        sys_prompt = 'You are an intelligent assistant who is good at mimicking human discussion in a multi-user collaboration.\n\n'
        input_prompt = user_persona_str+question_info_str
        message_sys = {"role": "system","content": sys_prompt}
        message_user = {"role": "user","content": input_prompt}
        message_list = [message_sys,message_user]
        response_str = self.llm_response_class.run_response(message_list,print_progress)
        if save_log_txt_path != None:
            if print_progress:
                self._store_log('\n\system prompt:'+sys_prompt,save_log_txt_path,color='blue',print=True)
                self._store_log('\n\ninput prompt:'+input_prompt,save_log_txt_path,color='red',print=True)
                self._store_log('\n\nllm response:'+response_str,save_log_txt_path,color='green',print=True)
            else:
                self._store_log('\n\system prompt:'+sys_prompt,save_log_txt_path,print=False)
                self._store_log('\n\ninput prompt:'+input_prompt,save_log_txt_path,print=False)
                self._store_log('\n\nllm response:'+response_str,save_log_txt_path,print=False)
        return response_str

    def step_forward():
        pass

    def summarize(self,idea_num,rank):
        pass 
        # if idea_num == 'all':

        # if rank == True:

    def _store_log(self,input_string,log_file,color=None, attrs=None, print=False):
        with open(log_file, 'a+') as f:
            f.write(input_string + '\n')
            f.flush()
        if(print):
            cprint(input_string, color=color, attrs=attrs)


class Multi_Agent_Brainstorm():
    def __init__(self,llm_response_class = None,openai_api_key = None,llm_model_name="gpt-3.5-turbo-0125",timeout=120,temperature=0,max_tokens=3000):
        '''
        The llm_response_class should at least have a run_response function to accept message list as input.
        Use .help() function to get more info about how to use this class.
        '''
        if llm_response_class == None:
            print('You did not set your own llm_response_class. We will use the default openai GPT 3.5 model. Please at least set your openai api_key.')
            self.llm_response_class = LLM_Respond_GPT_Class(openai_api_key=openai_api_key,model_name=llm_model_name,timeout=timeout,temperature=temperature,max_tokens=max_tokens)
        else:
            self.llm_response_class = llm_response_class
            print('Load your LLM successfully!')
    
    def help(self):
        print('\n\nNote that the instruction is very important for you to format the output of LLMs.')
        print('\n\nFor example, you can provide examples to format the exact output of LLMs so that it is easy for you to extract data from LLM answers automatically.')
        print('\n\nHere are example usage for your reference.\n\n')
        print('LLM_Respond_Class_Item = LLM_Respond_GPT_Class(openai_api_key='',model_name="gpt-3.5-turbo-0125",timeout=120,temperature=0,max_tokens=3000)')
        print('Survey_Pipeline_Demo = Multi_Agent_Brainstorm(llm_response_class = LLM_Respond_Class_Item)')
        print('Survey_Pipeline_Demo.init_agent(agent_num = 3, round_num = 3)')
        print('Survey_Pipeline_Demo.run_brainstorm(topic = \'your goal\',instruction = \'your output format\',conversation_json = \'conversation.json\',print_progress=True,save_log_txt_path=\'log.txt\')')

    def _store_log(self,input_string,log_file,color=None, attrs=None, print=False):
        with open(log_file, 'a+') as f:
            f.write(input_string + '\n')
            f.flush()
        if(print):
            cprint(input_string, color=color, attrs=attrs)
    
    def init_agent(self,agent_num,round_num):
        self.agent_num = agent_num
        self.round_num = round_num
        self.agent_dict = {f'agent_{int(i)}': Node_Agent(self.llm_response_class) for i in range(self.agent_num)}

    def run_brainstorm(self,topic,instruction,idea_num,rank,conversation_json,print_progress=False,save_log_txt_path=None):
        conversation_dict = {'topic':topic,'instruction':instruction}
        
        for round_id in range(self.round_num):
            conversation_dict[f'round_{int(round_id)}'] = {}
            for agent_id in range(self.agent_num):
                agent_response = self.agent_dict[f'agent_{int(agent_id)}'].run_respond(conversation_dict,print_progress,save_log_txt_path)
                conversation_dict[f'round_{int(round_id)}'][f'agent_{int(agent_id)}'] = agent_response

        with open(conversation_json, 'w') as f:
            json.dump(conversation_dict, f, indent=4)
        print('\nConversation_dict has been stored/saved in your local device.')





import re 
from termcolor import colored, cprint
from sklearn.metrics import accuracy_score, f1_score
import pandas as pd 

def calculate_metrics(result_path):
    data_result = pd.read_csv(result_path,sep='\t')

    label_list,pred_init_list,pred_post_list = [],[],[]
    student_list = list(set(data_result['student_id']))
    student_list.sort()
    for student_id in student_list:
        data_item = data_result[data_result['student_id']==student_id]
        future_label_str = data_item['future_label'].values[0]
        future_predict_post_str = data_item['future_predict_post'].values[0]
        future_predict_init_str = data_item['future_predict_init'].values[0]
        # Q9:1; Q10:0; Q11:0; Q12:0
        future_label_dict = convert_predict_string_to_dict(future_label_str)
        future_predict_post_dict = convert_predict_string_to_dict(future_predict_post_str)
        future_predict_init_dict = convert_predict_string_to_dict(future_predict_init_str)
        question_list = list(future_label_dict.keys())
        question_list.sort()
        for question_id in question_list:
            label_list.append(future_label_dict[question_id])
            pred_init_list.append(future_predict_init_dict[question_id])
            pred_post_list.append(future_predict_post_dict[question_id])

    accuracy_init = accuracy_score(label_list, pred_init_list)
    f1_init = f1_score(label_list, pred_init_list, average='weighted')
    accuracy_post = accuracy_score(label_list, pred_post_list)
    f1_post = f1_score(label_list, pred_post_list, average='weighted')
    print(f'accuracy_init: {accuracy_init}, accuracy_post: {accuracy_post}')
    print(f'f1_init: {f1_init}, f1_post: {f1_post}')


def convert_predict_string_to_dict(s):
    pairs = s.split("; ")
    result_dict = {}
    for pair in pairs:
        key, value = pair.split(":")
        key = int(key.replace("Q", ""))
        value = int(value)
        result_dict[key] = value
    
    return result_dict

def _store_log(input_string,log_file,color=None, attrs=None, print=False):
    with open(log_file, 'a+') as f:
        f.write(input_string + '\n')
        f.flush()
    if(print):
        cprint(input_string, color=color, attrs=attrs)

def dict_to_string(input_dict):
    return "; ".join(f"Question {key}: {value}" for key, value in input_dict.items())

def question_dict_to_string(input_dict):
    output = []
    for question, result in input_dict.items():
        assert result in [0,1]
        status = "Correct" if result == 1 else "Wrong"
        output.append(f"Question {question}: {status}")
    return "; ".join(output)

def question_dict_to_string_concise(input_dict):
    output = []
    for question, result in input_dict.items():
        assert result in [0,1]
        status = "1" if result == 1 else "0"
        output.append(f"Q{question}:{status}")
    return "; ".join(output)

def _generate_past_question_correctness_concise(data_table):
    question_id_list = list(set(data_table['question_id']))
    question_id_list.sort()
    return_str = ''
    for question_id in question_id_list:
        data_question = data_table[data_table['question_id']==question_id]
        assert len(data_question) == 1
        correctness = int(data_question['correctness'].values[0])
        assert correctness in [0,1]
        correctness_str = f'Question {question_id}: CORRECT; ' if correctness == 1 else f'Question {question_id}: Wrong; '
        return_str += correctness_str
    return return_str

def _generate_past_question_correctness_info(data_table,include_choice = True):
    question_id_list = list(set(data_table['question_id']))
    question_id_list.sort()
    question_str = '\nThe past questions are depicted below: \n'
    for question_id in question_id_list:
        data_question = data_table[data_table['question_id']==question_id]
        assert len(data_question) == 1
        question_content = data_question['question_content'].values[0]
        choice_content = data_question['choice_content'].values[0]
        course_content = data_question['course_content'].values[0]
        correctness = int(data_question['correctness'].values[0])
        assert correctness in [0,1]
        correctness_str = f'The student has a CORRECT answer for this question {question_id}.' if correctness == 1 else f'The student has a WRONG answer for this question {question_id}.'
        if include_choice == True:
            question_str += (
                f'\n Question {question_id}: {question_content}.\n'
                # +'Choices for this question: \n'
                +choice_content+'\n\n'
                # +'\n\n This question is related to course materials below: \n'
                # +course_content+'\n\n'
                +correctness_str+'\n'
                # +'-'*80+'\n\n'
            )
        else:
            question_str += (
                f'\n Question {question_id}: {question_content}.\n'
                +correctness_str+'\n'
            )
    return question_str

def _generate_future_question_info(data_table,include_choice = True):
    question_id_list = list(set(data_table['question_id']))
    question_id_list.sort()
    question_str = '\nThe future questions are depicted below: \n'
    for question_id in question_id_list:
        data_question = data_table[data_table['question_id']==question_id]
        assert len(data_question) == 1
        question_content = data_question['question_content'].values[0]
        choice_content = data_question['choice_content'].values[0]
        course_content = data_question['course_content'].values[0]
        if include_choice == True:
            question_str += (
                f'\n Question {question_id}: {question_content}.\n'
                # +'Choices for this question: \n'
                +choice_content+'\n\n'
                # +'\n\n This question is related to course materials below: \n'
                # +course_content+'\n'
                # +'-'*80+'\n\n'
            )
        else:
            question_str += (
                f'\n Question {question_id}: {question_content}.\n'
            )
    return question_str
        
def _generate_specific_eval_string(future_predict_dict,future_label_dict):
    assert len(future_predict_dict) == len(future_label_dict)
    specific_eval_str = 'Specifically, '
    question_id_list = list(future_predict_dict.keys())
    question_id_list.sort()
    for question_id in question_id_list:
        label_value = future_label_dict[question_id]
        pred_value = future_predict_dict[question_id]
        assert label_value in [0,1] and pred_value in [0,1]
        label = 'CORRECT' if label_value == 1 else 'WRONG'
        pred = 'CORRECT' if pred_value == 1 else 'WRONG'
        criteria = 'Correctly' if label == pred else 'Wrongly'
        if criteria == 'Wrongly':
            specific_eval_str += f'Question {question_id} is {criteria} predicted. Your prediction is that the student will have a {pred} answer. But actually, from the labels, the student has a {label} answer.\n'


    return specific_eval_str



def _extract_llm_predict_correctness(input_string,pattern = r"question\s*(\d+)\s*:\s*(correct|wrong)"):
    input_string_lower = input_string.lower()
    matches = re.findall(pattern, input_string_lower)
    print(matches)
    result_dict = {}
    # Create dictionary from matches
    for key, value in matches:
        assert value in ['correct','wrong']
        result_dict[int(key)] = 1 if value == 'correct' else 0
    
    return result_dict

def _extract_llm_predict_reason(input_string,pattern): 
    input_string_lower = input_string.lower()
    matches = re.findall(pattern, input_string_lower)
    print(matches)
    result_dict = {}
    # Create dictionary from matches
    for key, correctness, reason in matches:
        result_dict[int(key)] = reason.strip().replace('\n','. ').replace('\t','. ')
    
    return result_dict


import pandas as pd
import numpy as np 
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt

class Cluster_Text_Data():
    def __init__(self):
        pass 

    def help(self):
        help_str = '''
            dataframe = pd.read_csv('fake_data.csv',sep='\t')
            Cluster_Text_Data_Class = Cluster_Text_Data()
            Cluster_Text_Data_Class.load_process_dataset(dataframe,cluster_column='llm_response',stop_words='english')
            clustered_labels = Cluster_Text_Data_Class.run_cluster(n_clusters=5,model_name='kmeans',random_state=0)
            Cluster_Text_Data_Class.visual_cluster(vis_model='pca',savefig_path='temp.png')
            print('clustered_labels: ',clustered_labels)
        '''
        print(help_str)

    def load_process_dataset(self,dataframe,cluster_column,stop_words='english'):
        self.df = dataframe
        self.df['text'] = self.df[cluster_column].str.lower()

        vectorizer = TfidfVectorizer(stop_words=stop_words)
        self.X = vectorizer.fit_transform(self.df['text'])

    def run_cluster(self,n_clusters,model_name='kmeans',random_state=0):
        if model_name == 'kmeans':
            kmeans = KMeans(n_clusters=n_clusters, random_state=random_state).fit(self.X)
            self.df['cluster'] = kmeans.labels_
        else:
            raise ValueError('This cluster model is not supported yet!')
            
        return np.array(self.df['cluster'])

    def visual_cluster(self,vis_model='pca',savefig_path='temp.png'):
        if vis_model == 'pca':
            pca = PCA(n_components=2)
            principal_components = pca.fit_transform(self.X.toarray())
            df_pca = pd.DataFrame(data = principal_components, columns = ['principal component 1', 'principal component 2'])
            df_pca['cluster'] = self.df['cluster']

            plt.figure(figsize=(8,6))
            plt.scatter(df_pca['principal component 1'], df_pca['principal component 2'], c=df_pca['cluster'])
            plt.title('Text Clustering with KMeans')
            plt.xlabel('Principal Component 1')
            plt.ylabel('Principal Component 2')
            plt.savefig(savefig_path)
            plt.show()
        else:
            raise ValueError('This visualization model is not supported yet!')




import json,os
import numpy as np 
import pandas as pd
from termcolor import colored, cprint
from .llm_api import LLM_Respond_GPT_Class
from .demo import demo_survey, demo_user_persona

class Survey_Pipeline():
    def __init__(self,llm_response_class = None,openai_api_key = None,llm_model_name="gpt-3.5-turbo-0125",timeout=120,temperature=0,max_tokens=3000):
        '''
        The llm_response_class should at least have a run_response function to accept message list as input.
        Use .help() function to get more info about how to use this class.
        '''
        if llm_response_class == None:
            print('You did not set your own llm_response_class. We will use the default openai GPT 3.5 model. Please at least set your openai api_key.')
            self.llm_response_class = LLM_Respond_GPT_Class(openai_api_key=openai_api_key,model_name=llm_model_name,timeout=timeout,temperature=temperature,max_tokens=max_tokens)
        else:
            self.llm_response_class = llm_response_class
            print('Load your LLM successfully!')
        
    def help(self):
        self.load_default_user_persona()
        self.load_default_survey()
        print('\n\nNote that the instruction item in each question is very important for you to format the output of LLMs.')
        print('\n\nFor example, you can provide examples to format the exact output of LLMs so that it is easy for you to extract data from LLM answers automatically.')
        print('\n\nHere are example usage for your reference.\n\n')
        print('LLM_Respond_Class_Item = LLM_Respond_GPT_Class(openai_api_key='',model_name="gpt-3.5-turbo-0125",timeout=120,temperature=0,max_tokens=3000)')
        print('Survey_Pipeline_Demo = Survey_Pipeline(llm_response_class = LLM_Respond_Class_Item)')
        print('Survey_Pipeline_Demo.init_user(user_json_file = \'user_demo.json\')')
        print('Survey_Pipeline_Demo.init_survey(survey_json_file = \'survey_demo.json\')')
        print('Survey_Pipeline_Demo.run_survey(output_csv_path = \'survey_data_demo.csv\',repeat_num=2,save_log_txt_path=\'log.txt\',print_progress=True)')

    def load_json_file(self,file_name):
        # Get the directory of the current file
        current_dir = os.path.dirname(__file__)
        # Construct the path to the JSON file
        json_file_path = os.path.join(current_dir, file_name)
        
        # Open and read the JSON file
        with open(json_file_path, 'r') as file:
            data = json.load(file)
        
        return data

    def _store_log(self,input_string,log_file,color=None, attrs=None, print=False):
        with open(log_file, 'a+') as f:
            f.write(input_string + '\n')
            f.flush()
        if(print):
            cprint(input_string, color=color, attrs=attrs)

    def load_default_user_persona(self):
        user_persona_dict = demo_user_persona
        # user_persona_dict = self.load_json_file('user_demo.json')
        # with open('./user_demo.json', 'r') as f:
        #     user_persona_dict = json.load(f)
        print(f'Here is the default example user personas.\n\n{user_persona_dict}')
        return user_persona_dict

    def load_default_survey(self):
        survey_dict = demo_survey
        # survey_dict = self.load_json_file('survey_demo.json')
        # with open('./survey_demo.json', 'r') as f:
        #     survey_dict = json.load(f)
        print(f'Here is the default example survey.\n\n{survey_dict}')
        return survey_dict

    def generate_default_user_persona(self):
        user_persona_dict = {
            'u1': {'age': '24 years old', 'job': 'engineer'},
            'u2': {'education': 'high school'},
            'u3': {'name': 'Tom', 'major': 'Computer Science', 'degree': 'Master'},
        }
        print(f'Here is the default example user personas.\n\n{user_persona_dict}')
        with open('user_demo.json', 'w') as f:
            json.dump(user_persona_dict, f, indent=4)
        print('\nDemo user has been stored/saved in your local package.')

    def generate_default_survey(self):
        survey_dict = {
            'q1': {'type': 'text-entry', 'content': 'What is your name?', 'action': 'text entry', 'instruction': 'Output your answers with reasons within 20 words.'},
            'q2': {'type': 'single-choice', 'content': 'What is your favorite fruit?', 'action': 'A. apple, B. banana, C. lemon', 'instruction': 'Output your single choice directly without reasons. Your output format should exactly like: choice: [choice].'},
            'q3': {'type': 'multiple-choice', 'content': 'What do you eat everyday?', 'action': 'A. noodles, B. rice, C. meat', 'instruction': 'Output your multiple choices directly with reasons. Your output format should be exactly like: choice: [choice], reason: []. For example, if you want to select A and C, your output should be: choice: A,C, reason: [my reasons...].'},
        }
        print(f'Here is the default example survey.\n\n{survey_dict}')
        with open('survey_demo.json', 'w') as f:
            json.dump(survey_dict, f, indent=4)
        print('\nDemo survey has been stored/saved in your local package.')

        
    def init_user(self,user_json_file = None):
        '''
        The user_json file should include [user_id,user_persona]
        '''
        if user_json_file == None:
            self.user_persona_dict = self.load_default_user_persona()
            print('No init users. We will load default demo users.')
        else:
            with open(user_json_file, 'r') as f:
                self.user_persona_dict = json.load(f)
            print('Load your users successfully!')

    def init_survey(self,survey_json_file = None):
        '''
        Your survey format must be a json file including...
        '''
        if survey_json_file == None:
            self.survey_dict = self.load_default_survey()
            print('No init survey. We will load default demo survey.')
        else:
            with open(survey_json_file, 'r') as f:
                self.survey_dict = json.load(f)
            print('Load your survey successfully!')

    def _get_user_persona_str(self,user_persona_dict):
        user_persona_str = '\n\nHere is your persona information: '
        persona_type_list = list(user_persona_dict.keys())
        persona_type_list.sort()
        for persona_type in persona_type_list:
            persona_content = user_persona_dict[persona_type]
            user_persona_str += f'Your #{persona_type}# is {persona_content}. '
        user_persona_str += '\n\n'
        return user_persona_str

    def _get_question_info_str(self,question_info_dict):
        try:
            question_type = question_info_dict['type']
            question_content = question_info_dict['content']
            question_action = question_info_dict['action']
            question_instruction = question_info_dict['instruction']
        except:
            raise ValueError('Your survey json file must have type/content/action/instruction for each question. Please use Survey_Pipeline.help() to find an example of how to use it.')
        if question_type == 'single-choice': 
            question_action_str = f'You need to select only one choice from all choices below: {question_action}'
        elif question_type == 'multiple-choice': 
            question_action_str = f'You can select multiple choices from all choices below: {question_action}'
        else:
            question_action_str = f'The actions you can take for the question is: {question_action}. \n'
        question_info_str = (
            '\n\nHere is question that you need to answer: \n'
            +f'The #question type# is: {question_type}. \n'
            +f'The #question content# is: {question_content}. \n'
            +question_action_str
            +f'To answer the question and format your output, the #question instruction# is below: {question_instruction}. \n\n'
        )
        return question_info_str

    def generate_response(self,user_persona_dict,question_info_dict,save_log_txt_path,print_progress):
        user_persona_str = self._get_user_persona_str(user_persona_dict)
        question_info_str = self._get_question_info_str(question_info_dict)
        sys_prompt = 'You are an intelligent assistant who is good at mimicking humans with specific personas to answer contextual questions in a survey.\n\n'
        input_prompt = user_persona_str+question_info_str
        message_sys = {"role": "system","content": sys_prompt}
        message_user = {"role": "user","content": input_prompt}
        message_list = [message_sys,message_user]
        response_str = self.llm_response_class.run_response(message_list,print_progress)
        if save_log_txt_path != None:
            if print_progress:
                self._store_log('\n\system prompt:'+sys_prompt,save_log_txt_path,color='blue',print=True)
                self._store_log('\n\ninput prompt:'+input_prompt,save_log_txt_path,color='red',print=True)
                self._store_log('\n\nllm response:'+response_str,save_log_txt_path,color='green',print=True)
            else:
                self._store_log('\n\system prompt:'+sys_prompt,save_log_txt_path,print=False)
                self._store_log('\n\ninput prompt:'+input_prompt,save_log_txt_path,print=False)
                self._store_log('\n\nllm response:'+response_str,save_log_txt_path,print=False)
        return response_str


    def run_survey(self,output_csv_path,repeat_num = 1,save_log_txt_path = None,print_progress=False):
        survey_data_header = ['user_id','repeat_id','question_id','response']
        survey_data = []
        user_list = list(self.user_persona_dict.keys())
        user_list.sort()
        question_list = list(self.survey_dict.keys())
        question_list.sort()
        for user_id in user_list:
            user_each_dict = self.user_persona_dict[user_id]
            for r in range(repeat_num):
                for question_id in question_list:
                    question_each_dict = self.survey_dict[question_id]
                    response = self.generate_response(user_each_dict,question_each_dict,save_log_txt_path,print_progress)
                    survey_data.append([user_id,r,question_id,response])
                    if print_progress:
                        print(f'user {user_id} finished question {question_id} in repetition {r}.')

        survey_data = pd.DataFrame(np.array(survey_data),columns=survey_data_header)
        survey_data.to_csv(output_csv_path,index=False)

